/**
 * This file is part of ArmarX.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @author     Jianfeng Gao ( jianfeng dot gao at kit dot edu )
 * @date       2022
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */

#pragma once

#include <ArmarXCore/core/logging/Logging.h>
#include <ArmarXCore/interface/serialization/Eigen.h>
#include <ArmarXCore/util/CPPUtility/TripleBuffer.h>
#include "MP.h"


namespace armarx::control::common::mp
{
    struct KeypointsMPInput : virtual MPInput
    {
        Eigen::VectorXf keypointPosition;
        Eigen::VectorXf keypointVelocity;
        double deltaT;
    };

    struct KeypointsMPOutput : virtual MPOutput
    {
        Eigen::VectorXf keypointPosition;
        Eigen::VectorXf keypointVelocity;
    };

    using KeypointsMPInputPtr  = std::shared_ptr<KeypointsMPInput>;
    using KeypointsMPOutputPtr = std::shared_ptr<KeypointsMPOutput>;



    class KeypointsMP : virtual public MP
    {
    public:
        KeypointsMP(const MPConfig& c);
        ~KeypointsMP(){};

        void run(MPInputPtr, MPOutputPtr) override;
    };

    typedef std::shared_ptr<KeypointsMP> KeypointsMPPtr;

} /// namespace armarx::control::common::mp
