/**
 * This file is part of ArmarX.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @author     Jianfeng Gao ( jianfeng dot gao at kit dot edu )
 * @date       2022
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */


#pragma once

#include <mplib/representation/vmp/PrincipalComponentVMP.h>

#include <armarx/control/common/mp/MPPoolInterface.h>
#include "MP.h"


namespace armarx::control::common::mp
{
    BETTER_ENUM(mpClass, int, JSMP, JSVelMP, TSMP, TSVelMP, KeypointsMP)

    struct MPInputOutput
    {
        MPPtr mp;
        MPInputPtr input;
        MPOutputPtr output;
    };
    using MPInputOutputPtr = std::shared_ptr<MPInputOutput>;



    class MPPool:
            public virtual MPPoolInterface
    {
    public:
        MPPool(){};
        ~MPPool(){};

        void runMPs();
        void createMPs(const MP::MPListConfig& mpListConfig);
        void reconfigureMPs(const MP::MPListConfig& mpListConfig);

    public:
        /// Interface functions
        std::string getNames(const Ice::Current& iceCurrent = Ice::emptyCurrent) override;
        /// control
        void start(const std::string& mpName = "all",
                   const Ice::DoubleSeq& startVec = std::vector<double>(),
                   const Ice::DoubleSeq& goalVec = std::vector<double>(),
                   Ice::Double timeDuration = -1.0,
                   const Ice::Current& iceCurrent = Ice::emptyCurrent) override;

        void stop(const std::string &mpName = "all", const Ice::Current& iceCurrent = Ice::emptyCurrent) override;
        void pause(const std::string &mpName = "all", const Ice::Current& iceCurrent = Ice::emptyCurrent) override;
        void resume(const std::string &mpName = "all", const Ice::Current& iceCurrent = Ice::emptyCurrent) override;
        void reset(const std::string &mpName = "all", const Ice::Current& iceCurrent = Ice::emptyCurrent) override;
        bool isFinished(const std::string &mpName = "all", const Ice::Current& iceCurrent = Ice::emptyCurrent) override;

        /// setting
        void learnFromCSV(const Ice::StringSeq& fileNames = std::vector<std::string>(), const Ice::Current& iceCurrent = Ice::emptyCurrent) override;

        void setGoal(const Ice::DoubleSeq& goals, const Ice::Current& iceCurrent = Ice::emptyCurrent) override;
        void setStartAndGoal(const Ice::DoubleSeq& starts, const Ice::DoubleSeq& goals, const Ice::Current& iceCurrent = Ice::emptyCurrent) override;
        void setViaPoint(Ice::Double u, const Ice::DoubleSeq& viapoint, const Ice::Current& iceCurrent = Ice::emptyCurrent) override;
        void removeAllViaPoint(const Ice::Current& iceCurrent = Ice::emptyCurrent) override;

        /// serialze
        //        void setWeight(const Ice::DoubleSeqSeq& weights, const Ice::Current& iceCurrent = Ice::emptyCurrent) override;
        //        Ice::DoubleSeqSeq getWeight(const Ice::Current& iceCurrent = Ice::emptyCurrent) override;
        std::string serialize(const Ice::Current& iceCurrent = Ice::emptyCurrent) override;
        Ice::DoubleSeq deserialize(const std::string&, const Ice::Current& iceCurrent = Ice::emptyCurrent) override;

        ///
        Ice::Double getCanVal(const std::string& mpName, const Ice::Current& iceCurrent = Ice::emptyCurrent) override;
        bool getMPEnabled(const Ice::Current &) override;

    protected:
        std::map<std::string, MPInputOutput> mps;
    };

    using MPPoolPtr = std::shared_ptr<MPPool>;
} /// namespace armarx::control::common::mp
