#define BOOST_TEST_MODULE Armarx::Control::Hardwareconfig

#define ARMARX_BOOST_TEST

#include <Eigen/Core>
#include <Eigen/Geometry>

#include <ArmarXCore/core/logging/Logging.h>

#include <armarx/control/Test.h>
#include <armarx/control/ethercat/DataInterface.h>


using namespace armarx::control::ethercat;


constexpr float Precision = 0.0001;


BOOST_AUTO_TEST_CASE(modularConvertedValueConfig_ARMAR_DE_Cla1_L)
{
    ModularConvertedValue<std::uint32_t, std::int64_t> mcv;

    constexpr float zeroOffset = 611'565;
    constexpr float discontinuityOffset = 0;
    constexpr float maxValue = 1 << 20; // 2^20.
    constexpr bool isInverted = true;

    std::uint32_t encoderValue = 0;

    encoderValue = 675'657;
    mcv.init(&encoderValue, zeroOffset, discontinuityOffset, maxValue, isInverted);
    BOOST_CHECK_CLOSE(mcv.value, -1.4299936683818621, Precision);

    encoderValue = 572'982;
    mcv.init(&encoderValue, zeroOffset, discontinuityOffset, maxValue, isInverted);
    BOOST_CHECK_CLOSE(mcv.value, -0.8147535223031191, Precision);

    encoderValue = 436'801;
    mcv.init(&encoderValue, zeroOffset, discontinuityOffset, maxValue, isInverted);
    BOOST_CHECK_CLOSE(mcv.value, 0.00125834361506244, Precision);

    encoderValue = 308'468;
    mcv.init(&encoderValue, zeroOffset, discontinuityOffset, maxValue, isInverted);
    BOOST_CHECK_CLOSE(mcv.value, 0.7702441110046249, Precision);

    encoderValue = 198'364;
    mcv.init(&encoderValue, zeroOffset, discontinuityOffset, maxValue, isInverted);
    BOOST_CHECK_CLOSE(mcv.value, 1.4299996604943148, Precision);
}
