/*
 * This file is part of ArmarX.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @package    ...
 * @author     Jianfeng Gao ( jianfeng dot gao at kit dot edu )
 * @date       2021
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */

#pragma once

#include <armarx/control/common/mp/TSMP.h>
#include <armarx/control/common/mp/JSMP.h>
#include <armarx/control/common/mp/MPPool.h>
#include <armarx/control/njoint_controller/task_space/ImpedanceController.h>
#include <armarx/control/njoint_mp_controller/task_space/ControllerInterface.h>
#include <armarx/control/njoint_mp_controller/task_space/aron/NJointMPControllerConfig.aron.generated.h>


namespace armarx::control::njoint_mp_controller::task_space
{
    using namespace armarx::control::njoint_controller::task_space;
    using namespace armarx::control::common;

    struct TSImpedanceMPConfig:
            virtual public control_law::TaskspaceImpedanceController::Config
    {
        mp::MP::MPListConfig mpConfig;
    };

    /**
    * @defgroup Library-NJointTSImpedanceMPController NJointTSImpedanceMPController
    * @ingroup Library-RobotUnit-NJointControllers
    * A description of the library NJointTSImpedanceMPController.
    *
    * @class NJointTSImpedanceMPController
    * @ingroup Library-NJointTSImpedanceMPController
    * @brief Brief description of class NJointTSImpedanceMPController.
    *
    * Detailed description of class NJointTSImpedanceMPController.
    */
    class NJointTSImpedanceMPController :
            virtual public NJointTaskspaceImpedanceController
            , virtual public mp::MPPool
            , virtual public NJointTSImpedanceMPControllerInterface
    {
    public:
        using AronDTO = arondto::TSImpedanceMPConfig;
        using BO = TSImpedanceMPConfig;

        NJointTSImpedanceMPController(const RobotUnitPtr& robotUnit, const NJointControllerConfigPtr& config, const VirtualRobot::RobotPtr&);

        std::string getClassName(const Ice::Current&) const override;
        void updateConfig(const ::armarx::aron::data::dto::DictPtr& dto,
                          const Ice::Current& iceCurrent = Ice::emptyCurrent) override;
        void rtPreActivateController() override;
        void additionalTask() override;

    protected:
        TripleBuffer<BO> bufferUserToAdditionalTask;
        TripleBuffer<BO> bufferUserToRt;

    private:
        std::atomic<bool> isMPReady{false};
    };
}
